//
//  It is file 'crcstream.hpp'
//
//  Copyright (C) 1992, 1993 Serge Pachkovsky, ps@che.nsk.su
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 1, or (at your option)
//  any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
struct GenericCompressor {
    virtual		   ~GenericCompressor(){} ;
    virtual void	    WriteChar( BYTE c ) = 0 ;
    virtual unsigned long   GetMaxMemoryUsed( void ) = 0 ;
    virtual unsigned long   GetCompressedSize( void ) ;
    } ;

struct GenericDecompressor {
    virtual		   ~GenericDecompressor(){} ;
    virtual int 	    ReadChar( void ) = 0 ;
    } ;

struct CopyCompressor : GenericCompressor {
    FILE		   *out ;
			    CopyCompressor( int handle ) ;
    virtual		   ~CopyCompressor() ;
    virtual void	    WriteChar( BYTE c ) ;
    virtual unsigned long   GetMaxMemoryUsed( void ) ;
    } ;

struct TrueCompressor : GenericCompressor {
    Model2Compressor	    str ;
			    TrueCompressor( int handle, int order, unsigned long memory, int chars ) ;
    virtual		   ~TrueCompressor() ;
    virtual void	    WriteChar( BYTE c ) ;
    virtual unsigned long   GetMaxMemoryUsed( void ) ;
    virtual unsigned long   GetCompressedSize( void ) ;
    } ;

struct CopyDecompressor : GenericDecompressor {
    FILE		   *in ;
			    CopyDecompressor( int handle ) ;
    virtual		   ~CopyDecompressor() ;
    virtual int 	    ReadChar( void ) ;
    } ;

struct TrueDecompressor : GenericDecompressor {
    Model2Decompressor	    str ;
			    TrueDecompressor( int handle, int order, unsigned long memory, int chars ) ;
    virtual		   ~TrueDecompressor() ;
    virtual int 	    ReadChar( void ) ;
    } ;

struct CompStream {
    GenericCompressor	   *str ;
    CRC_32		    crc ;
			    CompStream( int handle, int order, unsigned long memory, int chars ) ;
			   ~CompStream() ;
    void		    WriteByte( BYTE c ) ;
    void		    WriteBuf( int size, BYTE * ) ;
    unsigned long	    GetCompressedSize( void ) ;
    } ;

struct DecompStream {
    GenericDecompressor    *str ;
    CRC_32		    crc ;
			    DecompStream( int handle, int order, unsigned long memory, int chars ) ;
			   ~DecompStream() ;
    int 		    ReadByte() ;
    void		    ReadBuf( int size, BYTE * ) ;
    } ;

inline unsigned long
GenericCompressor::GetCompressedSize( void )
{
return 0 ;
}

inline
TrueCompressor::TrueCompressor( int handle, int order, unsigned long memory, int chars ) :
    str( handle, order, memory, chars )
{
}

inline
TrueCompressor::~TrueCompressor()
{
}

inline void
TrueCompressor::WriteChar( BYTE c )
{
str.WriteChar( c ) ;
}

inline unsigned long
TrueCompressor::GetMaxMemoryUsed()
{
return str.GetMaxMemoryUsed() ;
}

inline unsigned long
TrueCompressor::GetCompressedSize()
{
return str.GetStreamSize() ;
}

inline
TrueDecompressor::TrueDecompressor( int handle, int order, unsigned long memory, int chars ) :
    str( handle, order, memory, chars )
{
}

inline
TrueDecompressor::~TrueDecompressor()
{
}

inline int
TrueDecompressor::ReadChar()
{
return str.ReadChar() ;
}

inline
CopyCompressor::CopyCompressor( int _handle )
{
    int     handle = dup( _handle ) ;

if( handle == -1 ){
    perror( "Duplicating handle in copy compressor" ) ;
    exit( -1 ) ;
    }
//
//  Silly BC 3.1 needs this conversion - I would not dream to do it myself...
//
if( ( out = fdopen( handle, (char *)ModeBinWrite ) ) == NULL ){
    cerr << "Can't initialize copy compressor\n" ;
    exit( -1 ) ;
    }
if( setvbuf( out, NULL, _IOFBF, IOBufSize ) != 0 ){
    cerr << "No memory for copy compressor's buffer\n" ;
    exit( -1 ) ;
    }
}

inline
CopyCompressor::~CopyCompressor()
{
if( out != NULL ){
    fclose( out ) ;
    out = NULL ;
    }
}

inline void
CopyCompressor::WriteChar( BYTE c )
{
putc( c, out ) ;
}

inline unsigned long
CopyCompressor::GetMaxMemoryUsed()
{
return 0 ;
}

inline
CopyDecompressor::CopyDecompressor( int _handle )
{
    int     handle = dup( _handle ) ;

if( handle == -1 ){
    perror( "Duplicating handle in copy decompressor" ) ;
    exit( -1 ) ;
    }
//
//  Silly BC 3.1 needs this conversion - I would not dream to do it myself...
//
if( ( in = fdopen( handle, (char *)ModeBinRead ) ) == NULL ){
    cerr << "Can't initialize copy decompressor\n" ;
    exit( -1 ) ;
    }
if( setvbuf( in, NULL, _IOFBF, IOBufSize ) != 0 ){
    cerr << "No memory for copy decompressor's buffer\n" ;
    exit( -1 ) ;
    }
}

inline
CopyDecompressor::~CopyDecompressor()
{
if( in != NULL ){
    fclose( in ) ;
    in = NULL ;
    }
}

inline int
CopyDecompressor::ReadChar()
{
return getc( in ) ;
}


#ifdef GEN_HUGE_FUNCS
HugeInline
CompStream::CompStream( int handle, int order, unsigned long memory, int chars )
{
if( (unsigned short) order == FileIsStored )
     str = new CopyCompressor( handle ) ;
else str = new TrueCompressor( handle, order, memory, chars ) ;
}

HugeInline
CompStream::~CompStream()
{
delete str ; str = NULL ;
}
#endif

inline void
CompStream::WriteByte( BYTE c )
{
str->WriteChar( c ) ;
crc.AddByte( c ) ;
}

inline void
CompStream::WriteBuf( int size, BYTE *p )
{
while( size-- > 0 ) WriteByte( *p++ ) ;
}

inline unsigned long
CompStream::GetCompressedSize()
{
return str->GetCompressedSize() ;
}

#ifdef GEN_HUGE_FUNCS
HugeInline
DecompStream::DecompStream( int handle, int order, unsigned long memory, int chars )
{
if( (unsigned short) order == FileIsStored )
     str = new CopyDecompressor( handle ) ;
else str = new TrueDecompressor( handle, order, memory, chars ) ;
}

HugeInline
DecompStream::~DecompStream()
{
delete str ; str = NULL ;
}
#endif

inline int
DecompStream::ReadByte()
{
    int     c = str->ReadChar() ;

if( c != EOF )
    crc.AddByte( c ) ;

return c ;
}

inline void
DecompStream::ReadBuf( int size, BYTE *p )
{
while( size-- > 0 ) *p++ = ReadByte() ;
}

