//
//  It is file 'namelist.cc'
//
//  Copyright (C) 1992, 1993 Serge Pachkovsky, ps@che.nsk.su
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 1, or (at your option)
//  any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
#include "stdinc.hpp"
#include "xio.hpp"
#include "namelist.hpp"

FileEntry::FileEntry( const char * const n )
{
next	  = NULL ;
name	  = strdup( n ) ;
CanonicalizeName( name ) ;
}

FileEntry::~FileEntry()
{
delete name ;
}

int
FileEntry::Match( const char * n, int recursive ) const
{
return patmat( n, name, recursive ) ;
}

FileList::FileList()
{
recursive = 0 ;
first	  = NULL ;
last	  = &first ;
}

FileList::~FileList()
{
    FileEntry	*p, *n ;

for( p = first ; p != NULL ; p = n ){
    n = p->next ;
    delete p ;
    }
}

void
FileList::Add( const char * const p )
{
    FileEntry	*e = new FileEntry( p ) ;

*last = e ;
last  = &( e->next ) ;
}

int
FileList::InList( const char * name ) const
{
    FileEntry	*p ;

for( p = first ; p != NULL ; p = p->next )
    if( p->Match( name, recursive ) )
	return 1 ;
return 0 ;
}

int
FileList::Empty()
{
return first == NULL ;
}

void
FileList::Recursive()
{
recursive = 1 ;
}

void
NameList::Recursive()
{
includes.Recursive() ;
}

void
NameList::AddName( const char * const name )
{
includes.Add( name ) ;
}

void
NameList::AddNamesFile( const char * const file )
{
    FILE    *in = fopen( file, "rt" ) ;

if( in == NULL ){
    perror( file ) ;
    return ;
    }

    char    buf[ 256 ] ;

while( fgets( buf, sizeof buf - 1, in ) != NULL ){
	int	size = strlen( buf ) ;
	char	*p = buf + size - 1 ;

    for( ; size > 0 && isspace( *p ) ; size--, p-- ) ;
    p[1] = 0 ;
    AddName( buf ) ;
    }
}

void
NameList::ExcludeName( const char * const name )
{
excludes.Add( name ) ;
}

int
NameList::InList( const char * name ) const
{
return includes.InList( name ) && ! excludes.InList( name ) ;
}

NameListIterator::NameListIterator( NameList &names )
{
list	      = &names ;
current_entry = list->includes.first ;
generator     = NULL ;
InitGenerator() ;
}

NameListIterator::~NameListIterator()
{
delete generator ;
}

int
NameListIterator::HaveNames()
{
return generator != NULL ;
}

void
NameListIterator::Next()
{
do {
    if( generator->Next() == -1 ){
	current_entry = current_entry->next ;
	InitGenerator() ;
	if( generator == NULL )
	    return ;
	}
    } while( list->excludes.InList( generator->CurrentName() ) ) ;
}

char * const
NameListIterator::Name() const
{
return generator->CurrentName() ;
}

void
NameListIterator::InitGenerator()
{
do {
    delete generator ; generator = NULL ;
    if( current_entry == NULL )
	return ;
    generator	  = new DirGen( current_entry->name, list->includes.recursive ) ;
    if( generator->CurrentName() != NULL )
	return ;
    current_entry = current_entry->next ;
    } while( 1 ) ;
}
