//
//  It is file 'psa.cc'
//
//  Copyright (C) 1992, 1993 Serge Pachkovsky, ps@che.nsk.su
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 1, or (at your option)
//  any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
#include "stdinc.hpp"
#include "timestamp.hpp"
#include "byteorder.hpp"
#include "archdr.hpp"
#include "machine.hpp"
#include "errcodes.hpp"
#include "xio.hpp"
#include "namelist.hpp"
#include "archive.hpp"

enum ArchiverActions {
    DisplayHelpMessage,     AddFiles,		    DeleteFiles,
    ExtractFiles,	    ListFiles,		    TestArchive,
    ShowGnuLicense,	    SayCrypticMessage
    } ;

enum Boolean {
    False,		    True
    } ;

static	ArchiverActions     action_code     = DisplayHelpMessage ;
static	int		    model_order     = 5 ;
static	unsigned long	    model_memory    = 4096 * 1024L ;
static	Boolean 	    test_archive    = False ;
static	NameList	    file_names ;
static	PsaErrorCodes	    top_error_code  = PsaCompletedOk ;
static	char *		    archive_name    = NULL ;
static	Archive 	    input_archive, output_archive ;

static void
display_title( void )
{
cerr <<
"Pretty Simple Archiver 0.91alpha Copyright (C) 1992,1993 Serge Pachkovsky\n"
"PSA comes with ABSOLUTELY NO WARRANTY; for details type: `psa $'\n" ;
}

static void
display_help_message( void )
{
cout <<
"Usage: PSA <command> [-option ...] <archive[.psa]> [filename ...]\n"
#ifdef TEST_VERSION
"\n"
"  ** WARNING: This is a test version of PSA, be careful with it **\n"
#endif
"\n"
"  <command>\n"
"   a: Add/Update files     d: Delete files     e: Extract files (w/pathnames)\n"
"   l: List of files        t: Test integrity of the archive\n"
"   $: Show license info\n"
"  <option>\n"
"     o<n>   Model order (default is " << model_order << ")\n"
"     m<nnn> Memory used for compression model (default is " << model_memory / 1024L << "K)\n"
"            I hate to say it, but at least 3M are necessary to handle binary\n"
"            files, and 400K are barely adequate for plain English text\n"
"     r      Recursively collect files\n"
"     t      Test archive after adding files\n"
"     -      '-' as usual letter\n"
#if defined( MSDOS )
"  Note: PSA uses unix-style wildcards, so all files in directory is '*',\n"
"  not '*.*' (which matches all files with non-empty extension)\n"
#endif
#if defined( GNUDOS ) && defined( MSDOS )
"  Don't forget to set environment variable GO32=noglob before running PSA\n"
#endif
;
}

static void
display_gnu_license( void )
{
cout <<
"\n"
"   This program is free software; you can redistribute it and/or modify\n"
"   it under the terms of the GNU General Public License as published by\n"
"   the Free Software Foundation; either version 1, or (at your option)\n"
"   any later version.\n"
"\n"
"   This program is distributed in the hope that it will be useful,\n"
"   but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
"   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
"   GNU General Public License for more details.\n"
"\n"
"   You should have received a copy of the GNU General Public License\n"
"   along with this program; if not, write to the Free Software\n"
"   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.\n"
"\n"
"   If you feel absolutely desperate to contact me, you could do so\n"
"   by e-mailing to ps@che.nsk.su (Internet) or 2:5000/19 (FIDOnet),\n"
"   but I can't guarantee to answer you.\n" ;
}

static void
say_cryptic_message( void )
{
//
//  This function have nothing to do with archiving. But I hope you
//  will not be tempted to remove it (as well as '$&' command). It is
//  so small and really important to me... Programming is an act of
//  expressing one's feelings, isn't it?
//
    char    cryptic[] = { // This array is intentionally made automatic!
'\n' ^ 0xA5, 'D'  ^ 0xA5, 'e'  ^ 0xA5, 'v'  ^ 0xA5, 'o'  ^ 0xA5, 't'  ^ 0xA5,
'e'  ^ 0xA5, 'd'  ^ 0xA5, ' '  ^ 0xA5, 't'  ^ 0xA5, 'o'  ^ 0xA5, ' '  ^ 0xA5,
'T'  ^ 0xA5, 'a'  ^ 0xA5, 'n'  ^ 0xA5, 'j'  ^ 0xA5, 'a'  ^ 0xA5, ','  ^ 0xA5,
' '  ^ 0xA5, 'w'  ^ 0xA5, 'i'  ^ 0xA5, 't'  ^ 0xA5, 'h'  ^ 0xA5, ' '  ^ 0xA5,
'm'  ^ 0xA5, 'y'  ^ 0xA5, ' '  ^ 0xA5, 'l'  ^ 0xA5, 'o'  ^ 0xA5, 'v'  ^ 0xA5,
'e'  ^ 0xA5, '.'  ^ 0xA5, '.'  ^ 0xA5, '.'  ^ 0xA5, '\n' ^ 0xA5, '\0' ^ 0xA5 } ;
    char    *p = cryptic ;

while( *p != char( 0xA5 ) ) cerr << char( *p++ ^ 0xA5 ) ;
}

static char *
make_good_archive_name( char *name )
{
    int     size = strlen( name ) ;
    char    *p	 = name + size - 1 ;
    int     i ;

CanonicalizeName( name ) ;
for( i = 4 ; size > 0 && i > 0 && *p != DirectorySeparator ; size--, p--, i-- )
    if( *p == '.' )
	return name ;

    char    *new_name = new char[ strlen( name ) + strlen( DefaultArchiveExtension ) + 1 ] ;

strcpy( new_name, name ) ;
strcat( new_name, DefaultArchiveExtension ) ;
return new_name ;
}

static int
parce_arguments( int argc, char **argv )
{
//  task name was already stripped out
if( argc <= 0 ){
    //	Kludge - display help message if no arguments specified
    action_code = DisplayHelpMessage ;
    return 0 ;
    }
//
//  Parce command code
//
switch( argv[0][0] ){
    case 'h': case 'H':
    case '?':
	action_code = DisplayHelpMessage ;
	return 0 ; // No more useful info on command line
    case '$':
	if( argv[0][1] == '&' )
	     action_code = SayCrypticMessage ;	// Why not? It's my program, and
						// no one can deny me this right...
	else action_code = ShowGnuLicense ;
	return 0 ;
    case 'a': case 'A':
	action_code = AddFiles ;
	break ;
    case 'd': case 'D':
	action_code = DeleteFiles ;
	break ;
    case 'e': case 'E':
    case 'x': case 'X':
	action_code = ExtractFiles ;
	break ;
    case 'l': case 'L':
	action_code = ListFiles ;
	break ;
    case 't': case 'T':
	action_code = TestArchive ;
	break ;
    default:
	cerr << "Unrecognized command '" << argv[0] << "', type psa h to see syntax help\n" ;
	return -1 ;
    }
//
//  Parce command options
//
    char    SwitchChar = '-' ;

for( argc--, argv++ ; argc > 0 && argv[0][0] == SwitchChar ; argc--, argv++ )
    switch( argv[0][1] ){
	case 'o': case 'O':
	    if( ( model_order  = atoi( argv[0] + 2 ) ) <= 0 ){
		cerr << "Can't parce model order: '" << argv[0] << "'\n" ;
		return -1 ;
		}
	    break ;
	case 'm': case 'M':
	    if( ( model_memory = atol( argv[0] + 2 ) * 1024L ) <= 0 ){
		cerr << "Can't parce model memory size: '" << argv[0] << "'\n" ;
		return -1 ;
		}
	    #ifndef DEBUG
	    if( model_memory < 64L * 1024 ){
		model_memory = 64L * 1024 ;
		cerr << "Model memory increased to 64K\n" ;
		}
	    #endif
	    break ;
	case 'r': case 'R':
	    file_names.Recursive() ;
	    break ;
	case 't': case 'T':
	    test_archive = True ;
	    break ;
	case '-':
	    SwitchChar	 = 0 ;	  // Stop arguments processing
	    break ;
	default:
	    cerr << "Unrecognized command option '" << argv[0] << "', type psa h to see syntax help\n" ;
	    return -1 ;
	}
//
//  Get archive name
//
if( argc <= 0 ){
    cerr << "Archive name should be specified\n" ;
    return -1 ;
    }
archive_name = make_good_archive_name( argv[0] ) ;
argc-- ; argv++ ;
//
//  Now, parce filenames (if any).
//
    long    file_count = 0 ;

for( ; argc > 0 ; argc--, argv++, file_count++ )
    if( argv[0][0] == '@' )
	 file_names.AddNamesFile( *argv + 1 ) ;
    else file_names.AddName( *argv ) ;

if( file_count == 0 ){
    file_names.AddName( AllFiles ) ;
    if( action_code == ListFiles || action_code == ExtractFiles )
	file_names.Recursive() ;
    }

return 0 ;
}

static void
update_files( void )
{
//
//  First, copy files from input archive
//
for( ; ! input_archive.Eof() ; input_archive.NextEntry() ){
	FileInfo	    info ;

    input_archive.FillFileInfo( info ) ;
    if( file_names.InList( info.name ) && access( info.name, 4 /* Read access */ ) == 0 )
	continue ; // Drop this file - we will add it later.
    cout << "Copying " << info.name ;
    output_archive.CopyCurrentFragment( input_archive ) ;
    cout << "\n" ;
    output_archive.NextEntry() ;
    }
//
//  Second, add new files to it.
//
	NameListIterator    files( file_names ) ;
	int		    errors = 0 ;

for( ; files.HaveNames() ; files.Next() ){
    output_archive.CompressFile( files.Name(), model_order, model_memory ) ;
    if( test_archive )
	if( output_archive.TestCurrentFragment() != 0 ){
	    top_error_code = PsaErrorInArchive ;
	    errors++ ;
	    }
    output_archive.NextEntry() ;
    }
//
//  Ok, we did it.
//
if( errors > 0 )
    cout << errors << " file(s) contain error(s)\n" ;
}

static void
list_files( void )
{
FileInfo::Title( cout ) ;

    unsigned long   comp = 0, uncomp = 0, files = 0 ;

for( ; ! input_archive.Eof() ; input_archive.NextEntry() ){
	FileInfo    info ;

    input_archive.FillFileInfo( info ) ;
    if( file_names.InList( info.name ) ){
	cout << info << "\n" ;
	files++ ;
	comp   += info.CompSize ;
	uncomp += info.UncompSize ;
	}
    }
if( files == 0 )
     cout << "No files found\n" ;
else cout << "\nTotal " << comp << "(" << uncomp << ") bytes in " << files << " file(s)\n" ;
}

static void
extract_files( void )
{
    int     errors = 0 ;

for( ; ! input_archive.Eof() ; input_archive.NextEntry() ){
	FileInfo    info ;

    input_archive.FillFileInfo( info ) ;
    if( file_names.InList( info.name ) )
	if( input_archive.ExtractCurrentFragment() != 0 ){
	    errors++ ;
	    top_error_code = PsaErrorInArchive ;
	    }
    }
if( errors > 0 )
    cout << "Total " << errors << " files with errors\n" ;
}

static void
delete_files( void )
{
//
//  This one is simple - just omit unneeded files...
//
for( ; ! input_archive.Eof() ; input_archive.NextEntry() ){
	FileInfo	    info ;

    input_archive.FillFileInfo( info ) ;
    if( file_names.InList( info.name ) ){
	cout << "Deleting " << info.name ;
	continue ; // Drop this file
	}
    output_archive.CopyCurrentFragment( input_archive ) ;
    output_archive.NextEntry() ;
    }
}

static void
test_files( void )
{
    int     errors = 0 ;

for( ; ! input_archive.Eof() ; input_archive.NextEntry() ){
	FileInfo    info ;

    input_archive.FillFileInfo( info ) ;
    if( file_names.InList( info.name ) )
	if( input_archive.TestCurrentFragment() != 0 ){
	    top_error_code = PsaErrorInArchive ;
	    errors++ ;
	    }
    }
if( errors > 0 )
    cout << "Total " << errors << " files with errors\n" ;
}

void
abort_handler( int subcode )
{
cerr << "\nAborted by Ctrl-C\n" ;
exit( PsaAborted ) ;
}

#ifndef ALIGNED_MEMORY_ACCESS
void
bus_handler( int subcode )
{
cerr << "\nBus error detected. You have compiled PSA with unaligned memory\n"
        "access enabled, and your machine probably cant handle that. Please\n"
	"try to recompile with -DALIGNED_MEMORY_ACCESS\n" ;
exit( PsaAborted ) ;
}
#endif

int
main( int argc, char *argv[] )
{
display_title() ;
if( parce_arguments( argc - 1, argv + 1 ) < 0 )
    return PsaSyntaxError ;
#ifdef __GNUC__
signal( SIGINT, abort_handler ) ; // Assume what we can't fail here...
#else
if( signal( SIGINT, abort_handler ) == SIG_ERR ){
    cerr << "Can't install SIGINT handler!\n" ;
    return PsaInitFailure ;
    }
#endif
#ifndef ALIGNED_MEMORY_ACCESS
signal( SIGBUS, bus_handler ) ;
#endif
switch( action_code ){
    case DisplayHelpMessage:
	display_help_message() ;
	return PsaCompletedOk ;
    case ShowGnuLicense:
	display_gnu_license() ;
	return PsaCompletedOk ;
    case SayCrypticMessage:
	say_cryptic_message() ;
	return PsaCompletedOk ;
    default:;	// Rest of commands will be serviced in a regular fascion elsewhere
	break ;
    }
tzset() ;	// Initialize time zone: stat() works incorrectly without it
		// under go32.
if( input_archive.Open( archive_name ) == -1 && action_code != AddFiles ){
    input_archive.Perror() ;
    return PsaNoArchive ;
    }
else
    file_names.ExcludeName( archive_name ) ;
if( action_code == AddFiles || action_code == DeleteFiles ){
    if( output_archive.Create( archive_name ) == -1 ){
	output_archive.Perror( "Creating output archive" ) ;
	return PsaNoArchive ;
	}
    file_names.ExcludeName( output_archive.GetName() ) ;
    }
switch( action_code ){
    default:
	cerr << "Undefined action code" ;
	return PsaInternalError ;
    case AddFiles:
	update_files() ;
	break ;
    case ListFiles:
	list_files() ;
	break ;
    case ExtractFiles:
	extract_files() ;
	break ;
    case DeleteFiles:
	delete_files() ;
	break ;
    case TestArchive:
	test_files() ;
	break ;
    }
if( action_code == AddFiles || action_code == DeleteFiles ){
    if( top_error_code == PsaCompletedOk ){
	input_archive.Close() ;
	if( input_archive.Backup() == -1 ){
	    input_archive.Perror( "Backing up original archive" ) ;
	    output_archive.Delete() ;
	    return PsaFileSystemError ;
	    }
	output_archive.Close() ;
	if( output_archive.Rename( archive_name ) == -1 ){
	    output_archive.Perror( "Renaming new archive" ) ;
	    input_archive.Rename( archive_name ) ;
	    return PsaFileSystemError ;
	    }
	if( input_archive.Delete() == -1 ){
	    input_archive.Perror( "Deleting old archive" ) ;
	    return PsaFileSystemError ;
	    }
	}
    else {
	if( output_archive.Delete() == -1 ){
	    output_archive.Perror( "Deleting bad archive" ) ;
	    return PsaFileSystemError ;
	    }
	input_archive.Close() ;
	}
    }
else {
    input_archive.Close() ;
    }
return top_error_code ;
}
