//
//  It is file 'timestamp.hpp'
//
//  Copyright (C) 1992, 1993 Serge Pachkovsky, ps@che.nsk.su
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 1, or (at your option)
//  any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
    typedef unsigned long   TimeScale ;
    const   TimeScale	SecondsBase = 0,    MinutesBase = 0,
			HoursBase   = 0,    DaysBase	= 1,
			MonthsBase  = 1,    YearsBase	= 1970 ;

    const   TimeScale	SecondsCnt  = 60,   MinutesCnt	= 60,
			HoursCnt    = 24,   DaysCnt	= 31,
			MonthsCnt   = 12,   YearsCnt	= 130 ;

    const   TimeScale	SecondsMult = 1,
			MinutesMult = SecondsMult * SecondsCnt,
			HoursMult   = MinutesMult * MinutesCnt,
			DaysMult    = HoursMult   * HoursCnt,
			MonthsMult  = DaysMult	  * DaysCnt,
			YearsMult   = MonthsMult  * MonthsCnt ;

class TimeStamp {
    TimeScale	    mix ;   //	All date/time parameters mixed

#define GenGetFunc(Name) \
    int  Get##Name() \
    { return ( mix / Name##Mult ) % Name##Cnt + Name##Base ; }
#define Add(Name,value) ( ( ( (value) - Name##Base) % Name##Cnt ) * Name##Mult )
#define GenSetFunc(Name) \
    void Set##Name(int value) \
    { mix = mix - ( (mix / Name##Mult) % Name##Cnt ) * Name##Mult + Add(Name,value) ; }

	static TimeScale Combine( int year, int month, int day, int hour, int minute, int second ){
		return Add(Years,year) + Add(Months,month)   + Add(Days,day) +
		       Add(Hours,hour) + Add(Minutes,minute) + Add(Seconds,second) ;
		}
    public:
	TimeStamp( int year = YearsBase, int month  = MonthsBase,  int day    = DaysBase,
		   int hour = HoursBase, int minute = MinutesBase, int second = SecondsBase ){
		mix = Combine( year, month, day, hour, minute, second ) ;
		}
	TimeStamp( struct tm *t ){
		mix = Combine( t->tm_year + 1900, t->tm_mon + 1, t->tm_mday, t->tm_hour, t->tm_min, t->tm_sec ) ;
		}
	GenGetFunc(Seconds) ;
	GenSetFunc(Seconds) ;
	GenGetFunc(Minutes) ;
	GenSetFunc(Minutes) ;
	GenGetFunc(Hours  ) ;
	GenSetFunc(Hours  ) ;
	GenGetFunc(Days   ) ;
	GenSetFunc(Days   ) ;
	GenGetFunc(Months ) ;
	GenSetFunc(Months ) ;
	GenGetFunc(Years  ) ;
	GenSetFunc(Years  ) ;
	void AllAtOnce( struct tm *t ){
		memset( (void *)t, 0, sizeof( struct tm ) ) ;
		t->tm_year = GetYears() - 1900 ;
		t->tm_mon  = GetMonths() - 1 ;
		t->tm_mday = GetDays() ;
		t->tm_hour = GetHours() ;
		t->tm_min  = GetMinutes() ;
		t->tm_sec  = GetSeconds() ;
		}

#undef	GenGetFunc
#undef	GenSetFunc
#undef	Add
    } ;

ostream & operator << ( ostream &, TimeStamp ) ;
int	  GetTimeStamp( const char *, TimeStamp & ) ;
int	  SetTimeStamp( const char *, TimeStamp   ) ;
