//
//  It is file 'xio.cc'
//
//  Copyright (C) 1992, 1993 Serge Pachkovsky, ps@che.nsk.su
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 1, or (at your option)
//  any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
#include "stdinc.hpp"
#include "xio.hpp"

FILE *
xfopen( const char * const name, const char * const mode )
{
if( strchr( mode, 'r' ) != NULL )
    return fopen( name, mode ) ;
//
//  Opening file for write, could have to make few directories...
//
    FILE   *f ;
    char   *tname = new char[ strlen( name ) + 1 ] ;

strcpy( tname, name ) ;

    char   *dir   = strchr( tname, '/' ) ;

while( ( f = fopen( name, mode ) ) == NULL ){
    if( dir == NULL || errno != ENOFILE ){
	delete tname ;
	return NULL ;
	}
    *dir = 0 ;
#ifdef MSDOS
    mkdir( tname ) ;
#else
    mkdir( tname, 0777 ) ;
#endif
    *dir = '/' ;
    dir  = strchr( dir + 1, '/' ) ;
    }
delete tname ;
return f ;
}

void
CanonicalizeName( char *name )
{
#ifdef MSDOS
//
//  Need to convert pathnames into canonical form
//
for( char *p = name ; *p != 0 ; p++ )
    if( *p == '\\' ) *p = DirectorySeparator ;
    else	     *p = tolower( *p ) ;
#endif
}

static int
name_patmat( const char * raw, const char * pat )
{
for( ; ; raw++, pat++ )
    switch( *pat ){
	default:
	    if( *raw != *pat )
		return 0 ;
	    if( *pat == 0 )
		return 1 ;  // EOS in pattern, match found
	    break ;
	case '?':
	    if( *raw == 0 )
		return 0 ;  // EOS in raw string, no match
	    break ;
	case '*':
	    ++pat ;
	    //
	    //	Try to find a match with zero or more symbols
	    //	from raw string...
	    //
	    do	if( name_patmat( raw, pat ) )
		    return 1 ;
	    while( *raw++ != 0 ) ;
	    return 0 ;
	}
}

int
patmat( const char * raw, const char * pat, int recursive )
{
    const char *    name_part = strrchr( pat, DirectorySeparator ) ;

if( name_part == NULL )
     name_part = pat ;
else name_part++ ;

    int 	    dir_len = name_part - pat ;

if( dir_len != 0 ){
    //
    //	Match directory name, if present in pattern
    //
    if( memcmp( raw, pat, dir_len ) != 0 )
	return 0 ;  // Directory name don't match
    }
raw += dir_len ;

    const char *    raw_name = strrchr( raw, DirectorySeparator ) ;

if( raw_name != NULL ){
    if( ! recursive )
	return 0 ;  // File is from offspring directory and recursive
		    // matching was not enabled
    raw_name++ ;
    }
else
    raw_name = raw ;

return name_patmat( raw_name, name_part ) ;
}

DirGen::DirGen( const char * pattern, int rec )
{
//cerr << "DirGen( " << pattern << ", " << rec << ")\n" ; // ????
recursive = rec ;
dir	  = NULL ;
nested	  = NULL ;
name	  = NULL ;

    const char *name_part = strrchr( pattern, DirectorySeparator ) ;
    const char *open_dir ;

if( name_part == NULL ){
    //
    //	No directory name, use empty one.
    //
    open_dir	 = "." ;    // Kludge - DJGPP's opendir thinks that "" is root,
			    // not current directory...
    base	 = new char[ 1 ] ;
    base[ 0 ]	 = 0 ;
    name_pattern = new char[ strlen( pattern ) + 1 ] ;
    strcpy( name_pattern, pattern ) ;
    }
else {
    //
    //	Have base directory part, split pattern into directory/name
    //
    name_part++ ;
    open_dir	 =
    base	 = new char[ name_part - pattern + 1 ] ;
    memcpy( base, pattern, name_part - pattern ) ;
    base[ name_part - pattern ] = 0 ;
    name_pattern = new char[ strlen( name_part ) + 1 ] ;
    strcpy( name_pattern, name_part ) ;
    }

// ????
//cerr << "base         = " << base         << "\n"
//     << "name_pattern = " << name_pattern << "\n"
//     << "open_dir     = " << open_dir     << "\n" ;

dir	  = opendir( (char *)open_dir ) ; // Actually, opendir() does not modify it's
					  // argument - it's just incorrectly defined ;(
if( dir == NULL )
    perror( base ) ;
PullNextFileName() ;
}

DirGen::~DirGen()
{
delete base ;
delete name_pattern ;
if( dir != NULL )
    closedir( dir ) ;
delete nested ;
delete name ;
}

int
DirGen::Next()
{
PullNextFileName() ;
if( name == NULL )
    return -1 ;
return 0 ;
}

char * const
DirGen::CurrentName() const
{
if( nested != NULL )
    return nested->CurrentName() ;
return name ;
}

void
DirGen::PullNextFileName()
{
//
//  First, try to scan nested directory
//
if( nested != NULL ){
    if( nested->Next() != -1 )
	return ;
    delete nested ; nested = NULL ;
    }
//
//  Remove name of previous file
//
delete name ; name = NULL ;

    struct dirent  *ent ;

while( ( ent = readdir( dir ) ) != NULL ){
    if( strcmp( ent->d_name, "." ) == 0 || strcmp( ent->d_name, ".." ) == 0 )
	continue ;  // Ignore special directories

    CanonicalizeName( ent->d_name ) ;

    name = new char[ strlen( base ) + strlen( ent->d_name ) + 1 ] ;
    strcpy( name, base ) ;
    strcat( name, ent->d_name ) ;

	struct	stat	stats ;

    if( stat( name, &stats ) == -1 )
	perror( name ) ;
    else
    if( (stats.st_mode & S_IFDIR) && recursive ){
	//
	//  Recursive search - scan subdirectories.
	//
	    char       *sname = new char[ strlen( name ) + 1 + strlen( name_pattern ) + 1 ] ;
	    const char	dir_switch[] = { DirectorySeparator, 0 } ;

	strcpy( sname, name ) ;
	strcat( sname, dir_switch ) ;
	strcat( sname, name_pattern ) ;
	nested = new DirGen( sname, recursive ) ;
	delete sname ;
	if( nested->CurrentName() != NULL )
	    return ; // Found name in the nested directory
	delete nested ; nested = NULL ;
	}
    else
    if( stats.st_mode & S_IFREG ){
	//
	//  Ordinary file, try to match it against our pattern
	//
//	cerr << "Matching " << ent->d_name << " against " << name_pattern << "\n" ; // ????
	if( name_patmat( ent->d_name, name_pattern ) )
	    return ;
//	cerr << "Failed" << "\n" ; // ????
	}
    delete name ; name = NULL ;
    }
}

//
//  And, couple of non-portable functions: determine
//  size of memory available.
//
#if defined( __TURBOC__ ) && defined( MSDOS )
unsigned long
AvailableVirtualMemory( unsigned long need )
{
return coreleft() ;
}

unsigned long
AvailablePhysicalMemory( unsigned long need )
{
return coreleft() ;
}
#else
//
//  Generic memory stat routines. Not very useful, probably,
//  but acceptable for most virtual-memory environments.
//
unsigned long
AvailableVirtualMemory( unsigned long need )
{
return AvailablePhysicalMemory( need ) + 128 * 1024 * 1024 ;
}

unsigned long
AvailablePhysicalMemory( unsigned long need )
{
return need ;
}
#endif
